function UpgradesManager:aquire_mixed(id, loading, identifier)
	if not tweak_data.upgrades.definitions[id] then
		return
	end
	local upgrade = tweak_data.upgrades.definitions[id]

	if upgrade.dlc and not managers.dlc:is_dlc_unlocked(upgrade.dlc) then
		return
	end
	if not identifier then
		identifier = UpgradesManager.AQUIRE_STRINGS[1]
	end
	local identify_key = Idstring(identifier):key()
	self._global.aquired[id] = self._global.aquired[id] or {}
	self._global.aquired[id][identify_key] = identifier
	self:_aquire_upgrade_mixed(upgrade, id, loading)
	self:setup_current_weapon()
end

function UpgradesManager:_aquire_upgrade_mixed(upgrade, id, loading)
	if upgrade.category == "weapon" then
		self:_aquire_weapon(upgrade, id, loading)
	elseif upgrade.category == "feature" then
		self:_aquire_feature_mixed(upgrade, id, loading)
	elseif upgrade.category == "equipment" then
		self:_aquire_equipment(upgrade, id, loading)
	elseif upgrade.category == "equipment_upgrade" then
		self:_aquire_equipment_upgrade(upgrade, id, loading)
	elseif upgrade.category == "temporary" then
		self:_aquire_temporary(upgrade, id, loading)
	elseif upgrade.category == "cooldown" then
		self:_aquire_cooldown(upgrade, id, loading)
	elseif upgrade.category == "team" then
		self:_aquire_team(upgrade, id, loading)
	elseif upgrade.category == "armor" then
		self:_aquire_armor(upgrade, id, loading)
	elseif upgrade.category == "rep_upgrade" then
		self:_aquire_rep_upgrade(upgrade, id, loading)
	elseif upgrade.category == "melee_weapon" then
		self:_aquire_melee_weapon(upgrade, id, loading)
	elseif upgrade.category == "grenade" then
		self:_aquire_grenade(upgrade, id, loading)
	end
end

function UpgradesManager:_aquire_feature_mixed(feature)
	managers.player:aquire_upgrade_mixed(feature.upgrade)
end

function UpgradesManager:unaquire_mixed(id, identifier)
	if not tweak_data.upgrades.definitions[id] then
		return
	end
	if not identifier then
		identifier = UpgradesManager.AQUIRE_STRINGS[1]
	end
	local identify_key = Idstring(identifier):key()
	if not self._global.aquired[id] or not self._global.aquired[id][identify_key] then
		return
	end
	self._global.aquired[id][identify_key] = nil
	local count = 0
	for key, aquired in pairs(self._global.aquired[id]) do
		count = count + 1
	end
	if count == 0 then
		self._global.aquired[id] = nil
		local upgrade = tweak_data.upgrades.definitions[id]
		self:_unaquire_upgrade_mixed(upgrade, id)
	end
end

function UpgradesManager:_unaquire_upgrade_mixed(upgrade, id)
	if upgrade.category == "weapon" then
		self:_unaquire_weapon(upgrade, id)
	elseif upgrade.category == "feature" then
		self:_unaquire_feature_mixed(upgrade, id)
	elseif upgrade.category == "equipment" then
		self:_unaquire_equipment(upgrade, id)
	elseif upgrade.category == "equipment_upgrade" then
		self:_unaquire_equipment_upgrade(upgrade, id)
	elseif upgrade.category == "temporary" then
		self:_unaquire_temporary(upgrade, id)
	elseif upgrade.category == "cooldown" then
		self:_unaquire_cooldown(upgrade, id)
	elseif upgrade.category == "team" then
		self:_unaquire_team(upgrade, id)
	elseif upgrade.category == "armor" then
		self:_unaquire_armor(upgrade, id)
	elseif upgrade.category == "melee_weapon" then
		self:_unaquire_melee_weapon(upgrade, id)
	elseif upgrade.category == "grenade" then
		self:_unaquire_grenade(upgrade, id)
	end
end

function UpgradesManager:_unaquire_feature_mixed(feature)
	managers.player:unaquire_upgrade_mixed(feature.upgrade)
end